"""Symbols for entities in factorio.

The name of the symbol is the same as the internal name but with ``-``
converted to ``_``.  For example, to refer to an "electronic-circuit" use
`itm.electronic_circuit`.

This module also provides a few special items.

.. py:data:: _combined_research

Result of `rcp._combined_research <factoriocalc.rcp._combined_research>`.

.. py:data:: _military_research

Result of `rcp._military_research <factoriocalc.rcp._military_research>`.

.. py:data:: _production_research

Result of `rcp._production_research <factoriocalc.rcp._production_research>`.


"""

byName = {}

